import json
import urllib
import warnings

import ee
import geopandas as gpd
import numpy as np
import pandas as pd

import waterpyk.errors as err
from waterpyk.calcs import combine_bands, interp_daily

ee.Initialize()


def extract_urls(gage, **kwargs):
    """
    Return relevant urls for a USGS gage.

    Args:
        gage (:obj:`str` or :obj:`int`): USGS 8-number gage ID. If int, leading 0s will automatically be added.
        **flow_start_date (str, optional): default: '1980-10-01'
        **flow_end_date (str, optional): default: '2021-10-01'

    Returns:
        str, str, str, str: 4 strings with urls which (1) access basin lat/long geometry. (2) access geometry of flowline (i.e. rivers) lat/long geometry. (3) access basin metadata. (4) access basin discharge timeseries (between the dates of **kwargs).
    """
    gage = str(gage)

    # Raise errors
    if len(gage) > 8:
        raise err.GageTooLongError(
            f'Gage ID length is {len(gage)} and cannot be greater than 8.')

    # Make sure gage ID is in correct format without missing starting 0s
    if (len(gage) < 8):
        warnings.warn(
            f'WARNING: Gage ID length is {len(gage)}. Zeros will be added to begginning until length = 8.')
        num = 8 - len(gage)
        gage = '0' * num + gage

    # Default start_date and end_date kwargs
    default_kwargs = {
        'flow_start_date': '1980-10-01',
        'flow_end_date': '2021-10-01',
    }
    kwargs = {**default_kwargs, **kwargs}

    # Data URLs
    url_basin_geometry = 'https://labs.waterdata.usgs.gov/api/nldi/linked-data/nwissite/USGS-%s/basin?f=json' % gage
    url_flow_geometry = 'https://labs.waterdata.usgs.gov/api/nldi/linked-data/nwissite/USGS-%s/navigation/UM/flowlines?f=json&distance=1000' % gage
    url_metadata = "https://labs.waterdata.usgs.gov/api/nldi/linked-data/nwissite/USGS-%s/?f=json" % gage
    url_flow = 'https://waterdata.usgs.gov/nwis/dv?cb_00060=on&format=rdb&site_no=' + gage + \
        '&referred_module=sw&period=&begin_date=' + \
        kwargs['flow_start_date'] + '&end_date=' + kwargs['flow_end_date']

    return url_basin_geometry, url_flow_geometry, url_metadata, url_flow


def extract_geometry(gage, **kwargs):
    """
    Get the geometry of a USGS gage in Google Earth Engine (GEE) form and as a geopandas dataframe.

    Args:
        gage (str or int): USGS 8-number gage ID. If int, leading 0s will automatically be added.

    Returns:
        :obj:`feature` and  :obj:`gdf`: GEE feature containing the basin's exterior polygon coordinates and geopandas dataframe containing the basin's coordinates.
    """
    urls = extract_urls(gage, **kwargs)
    # Access site geometry
    basin_geometry = gpd.read_file(urls[0])
    poly_coords = [item for item in basin_geometry.geometry[0].exterior.coords]
    # , {'Name': str(site_name[0]), 'Gage':int(watershed)})
    gee_feature = ee.Feature(ee.Geometry.Polygon(coords=poly_coords))

    return gee_feature, basin_geometry


def extract_metadata(gage, **kwargs):
    """
    Get metadata for a USGS gage.

    Args:
        gage (str or int): USGS 8-number gage ID. If int, leading 0s will automatically be added.

    Returns:
        str, str: 2 strings. (1) USGS long-name of gage. (2) description of the form 'USGS Basin + gage ID + imported at + site_name + CRS: + coordinate system
    """
    url_basin_geometry, url_flow_geometry, url_metadata, url_flow = extract_urls(
        gage, **kwargs)
    request_metadata = urllib.request.urlopen(url_metadata)  # type: ignore
    basin_geometry = gpd.read_file(url_basin_geometry)  # for CRS
    site_name = [json.load(request_metadata)['features']
                 [0]['properties']['name'].title()]
    description = 'USGS Basin (' + str(gage) + ') imported at ' + \
        str(site_name[0]) + 'CRS: ' + str(basin_geometry.crs)

    return site_name, description


def extract_streamflow(gage, **kwargs):
    """
    Get a dataframe with streamflow (i.e. discharge) for a USGS gage. Units are converted to mm using the area of the basin, calculated from the exterior geometry.

    Args:
        gage (str or int): USGS 8-number gage ID. If int, leading 0s will automatically be added.
        **flow_start_date (str, optional): default: '1980-10-01'
        **flow_end_date (str, optional): default: '2021-10-01'

    Returns:
        :obj:`df`: dataframe with daily discharge (Q) in units of cfs, m3/day, m, and mm.
    """
    # get URLs and basin geometry
    url_basin_geometry, url_flow_geometry, url_metadata, url_flow = extract_urls(
        gage, **kwargs)
    print('\nStreamflow data is being retrieved from:', url_flow_geometry, '\n')
    basin_geometry = gpd.read_file(url_basin_geometry)  # for drainage area
    drainage_area_m2 = basin_geometry.to_crs('epsg:26910').geometry.area

    # read streamflow data and clean csv
    df = pd.read_csv(url_flow, header=31, delim_whitespace=True)
    df.columns = ['usgs', 'site_number', 'datetime', 'Q_cfs', 'a']
    df['date'] = pd.to_datetime(df.datetime)
    df = df[['Q_cfs', 'date']]

    # Convert Q to m^2 using drainage area
    # this is needed because sometimes there are non-numeric entries and we want to ignore them
    df['Q_cfs'] = df['Q_cfs'].astype(float, errors='ignore')
    df['Q_m3day'] = (86400*df['Q_cfs'])/(35.31)  # m3/day
    df['Q_m'] = df['Q_m3day'] / float(drainage_area_m2)
    df['Q_mm'] = df['Q_m3day'] / float(drainage_area_m2) * 1000

    return df


def extract_geometry_flowline(gage, **kwargs):
    """
    Get geopandas dataframe of USGS basin flowline geometry for plotting.

    Args:
        gage (str or int): USGS 8-number gage ID. If int, leading 0s will automatically be added.

    Returns:
        :obj:`df`: geopandas dataframe with geometry of flowlines (rivers) for plotting.
    """
    urls = extract_urls(gage, **kwargs)
    geometry_df = gpd.read_file(urls[1])

    return geometry_df


def extract_latitude(gage, **kwargs):
    """
    Get the latitude of the centroid of the USGS gage for calculating PET.

    Args:
        gage (str or int): USGS 8-number gage ID. If int, leading 0s will automatically be added.

    Returns:
        float: latitude
    """
    urls = extract_urls(gage, **kwargs)
    basin_geometry = gpd.read_file(urls[0])
    latitude = basin_geometry.to_crs('epsg:4326').geometry[0].centroid.y

    return latitude




   
"""Top-level package for waterpyk."""

__author__ = """Erica McCormick"""
__email__ = "erica.elmstead@gmail.com"
__version__ = "1.1.1"

import sys
import os
import pkg_resources
import pandas as pd

def load_data(layers):
    """Finds the import data (either all.csv or minimal.csv corresponding to layers input) and loads it.
    
    Args:
        layers (str, optional): the GEE input layers to be extracted. Options if string are 'minimal' or 'all'.
    
    Returns:
        :obj:`df`: dataframe of read csv from layers.
    """
    
    file_path = 'layers_data/' + layers + '.csv'
    stream = pkg_resources.resource_stream(__name__, file_path)
    return pd.read_csv(stream, encoding='latin-1')

def in_colab_shell():
    """Checks if code is being executed within Google Colab to decide on default_saving_dir (default saving directory path).
    Function from geemap by Qiusheng Wu
    
    Returns:
        bool
    """

    if "google.colab" in sys.modules:
        return True
    else:
        return False

if in_colab_shell():
    print('\nOutput will be saved at /content/drive/MyDrive' + '\n')
    default_saving_dir = '/content/drive/MyDrive'
else:
    path = os.path.join(os.getcwd(), 'data', 'output')
    if os.path.exists(path):
        print('\nOutput will be saved at existing directory' + path + '\n')
        default_saving_dir = path
    else:
        print('\nMaking new directory for output at' + path + '\n')
        os.mkdir(os.path.join(os.getcwd(), 'data'))
        os.mkdir(os.path.join(os.getcwd(), 'data', 'output'))
        default_saving_dir = path
        
        
    


import ee
import geopandas as gdp
from waterpyk import gee

ee.Initialize()


def test_checks_if_gdf_to_feat_returns_feat():
    gdf = gdp.read_file('tests/testing_data/test_json_sanbern.json')
    gee_feat = gee.gdf_to_feat(gdf)
    assert type(gee_feat) is ee.Feature


def test_confirms_target_epsg_for_gdf_to_feat():
    gdf = gdp.read_file('tests/testing_data/test_json_sanbern.json')
    gee_feat = gee.gdf_to_feat(gdf)
    gee_feat.projection().getInfo()


def test_extract_basic_prism_for_all_reducers():
    gdf = gdp.read_file('tests/testing_data/test_json_sanbern.json')
    gee_feature = gee.gdf_to_feat(gdf)
    kind = 'watershed'
    asset_id = "OREGONSTATE/PRISM/AN81m"
    scale = 500
    bands = 'ppt'
    relative_date = None
    bands_to_scale = None
    scaling_factor = 1
    reducer_type = None
    new_bandnames = None
    interp = False

    start_date = '2005-10'
    end_date = '2005-12'
    df = gee.extract_basic(gee_feature, kind, asset_id, scale, bands, start_date, end_date,
                           relative_date, bands_to_scale, scaling_factor, reducer_type, new_bandnames, interp)
    assert round(df['value'].sum(
    ), 3) == 48.035, "PRISM summed for 2 months without interpolation changed"

    start_date = '2005-10'
    end_date = '2006-10'
    interp = True
    df = gee.extract_basic(gee_feature, kind, asset_id, scale, bands, start_date, end_date,
                           relative_date, bands_to_scale, scaling_factor, reducer_type, new_bandnames, interp)
    assert round(df['value'].sum(
    ), 3) == 175343.377, "PRISM summed for 1 year with interpolation changed"
    assert str(df['date'][0]) == '2001-06-20 00:00:00'

    interp = False
    df = gee.extract_basic(gee_feature, kind, asset_id, scale, bands, start_date, end_date,
                           relative_date, bands_to_scale, scaling_factor, reducer_type, new_bandnames, interp)
    assert str(df['date'][0]) == '2001-05'



import geopandas as gpd
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from waterpyk import errors, watershed

# python3 -m pytest tests -v --disable-warnings


def test_should_always_pass():
    assert 2 + 2 == 4, "This is just a dummy test"

# def test_should_always_fail():
#    assert 2 + 2 == 6, "This is just a dummy test"

############### TESTING EXTRACT_URLS() ###############


@pytest.fixture
def usgs_gauge():
    gage = 11475560  # Elder Creek
    return gage


def test_that_default_kwargs_have_not_changed():
    urls = watershed.extract_urls(11475560)
    assert '1980-10-01' in urls[3], "flow_start_date default_kwarg has changed"
    assert '2021-10-01' in urls[3], "flow_end_date default_kwarg has changed"


def test_should_find_four_urls_for_str_or_int_gaugeid():
    urls = watershed.extract_urls(11475560)
    urls_str = watershed.extract_urls('11475560')
    assert len(urls) == 4, "four urls were not returned using int gauge ID"
    assert len(urls_str) == 4, "four urls were not returned using str gauge ID"


def test_checks_urls_in_expected_order():
    urls = watershed.extract_urls(11475560)
    assert 'basin' in urls[0], "the first url does not include the str 'basin'"
    assert 'flowlines' in urls[1], "the second url does not include the str 'flowline'"
    assert '/?f=json' in urls[2], "the third url does not have the json identifier immediately after the slash"
    assert 'begin_date' in urls[3], "the fourth url does not have the str 'begin_date'"


def test_checks_geometry_col_exists_in_basin_and_flow_gdfs():
    urls = watershed.extract_urls(11475560)
    basin_geometry = gpd.read_file(urls[0])
    flow_geometry = gpd.read_file(urls[1])
    assert 'geometry' in list(basin_geometry.columns)
    assert 'geometry' in list(flow_geometry.columns)


def test_should_raise_error_if_gauge_id_too_long():
    with pytest.raises(errors.GageTooLongError) as exception_info:
        urls = watershed.extract_urls(gage=1111111111)
    assert exception_info.value.args[0] == 'Gage ID length is 10 and cannot be greater than 8.'


def test_preceding_zeros_added_for_gaugeid_too_short():
    urls_int_gauge = watershed.extract_urls(gage=1234)
    urls_str_gauge = watershed.extract_urls(gage='1234')
    expected_gauge_id = '00001234'
    assert expected_gauge_id in urls_int_gauge[0]
    assert expected_gauge_id in urls_str_gauge[0]


def test_if_kwargs_changes_work_throughout():
    begin_date = '1985-10-01'
    end_date = '1986-10-01'
    urls = watershed.extract_urls(11475560,
                                  flow_start_date=begin_date,
                                  flow_end_date=end_date)
    # Check that begin_date is in the urls
    assert begin_date in urls[3], "begin date is not in flow URL"
    assert end_date in urls[3], "end date is not in flow URL"
    df = watershed.extract_streamflow(11475560,
                                      flow_start_date=begin_date,
                                      flow_end_date=end_date)
    assert df.iloc[0]['date'].strftime(
        '%Y-%m-%d') >= begin_date, "data exists before begin_date"
    assert df.iloc[-1]['date'].strftime(
        '%Y-%m-%d') <= end_date, "data exists after end_date"


def test_checks_if_usgs_api_streamflow_has_changed():
    # Get new discharge csv using the same (default) dates
    kwargs = {'flow_start_date': '1980-10-01',
              'flow_end_date': '2021-10-01'}
    df_new = watershed.extract_streamflow(11475560, **kwargs)
    df_new.to_csv('new_gauge.csv')
    # Pull in csv of old version
    df_old = pd.read_csv('tests/testing_data/flow_11475560.csv')
    df_new = df_new[['Q_cfs', 'Q_mm']]
    df_old = df_old[['Q_cfs', 'Q_mm']]
    assert_frame_equal(
        df_old, df_new, check_dtype=False), "expected Elder Q has changed or other error has occured"


def test_checks_if_latitude_changed():
    lat = watershed.extract_latitude(11475560)
    assert round(lat, 5) == 39.71395
